package edu.kit.ipd.descartes.adaptation;

import java.io.FileInputStream;
import java.io.IOException;
import java.util.Properties;

import org.apache.log4j.BasicConfigurator;
import org.apache.log4j.Level;
import org.apache.log4j.Logger;
import org.eclipse.emf.common.util.URI;
import org.eclipse.emf.ecore.xmi.DanglingHREFException;

import edu.kit.ipd.descartes.adaptation.evaluation.WeightingFunctionEvaluator;
import edu.kit.ipd.descartes.adaptation.event.EventTypeEnum;
import edu.kit.ipd.descartes.adaptation.model.analysis.IModelAnalyzer;
import edu.kit.ipd.descartes.adaptation.model.analysis.pcm.PcmModelAnalyzer;
import edu.kit.ipd.descartes.adaptation.model.dmm.AdaptationProcessModelLoader;
import edu.kit.ipd.descartes.adaptation.model.dmm.DmmModelActionHandler;
import edu.kit.ipd.descartes.adaptation.model.dmm.util.ObjectivesHelper;
import edu.kit.ipd.descartes.adaptation.model.perfdatarepo.IPerformanceDataRepositoryHandler;
import edu.kit.ipd.descartes.adaptation.model.perfdatarepo.PerformanceDataRepositoryHandlerFactory;
import edu.kit.ipd.descartes.mm.adaptation.AdaptationProcess;
import edu.kit.ipd.descartes.mm.adaptation.Event;
import edu.kit.ipd.descartes.mm.adaptation.Strategy;
import edu.kit.ipd.descartes.mm.adaptation.WeightedTactic;
import edu.kit.ipd.descartes.perfdatarepo.Impact;

public class AdaptationControl {

    private static final String DEFAULT_PROP_FILE_PATH = "./default.properties";
    private static final String PROP_ADAPTATION_PROCESS_XMI_FILE_PATH_NAME = "adaptationprocess.instance";
    private static final String PROP_PERF_DATA_REPO_XMI_FILE_PATH_NAME = "perfdatarepo.instance";
    private static final String PROP_PERF_DATA_REPO_METAMODEL_TYPE = "perfdatarepo.metamodel";
    private static final String PROP_ADAPTATION_PROCESS_NAME = "process.name";
    private static final String PROP_MAX_ITERATIONS = "process.maxiterations";
    private static final String PROP_TRIGGERING_EVENT_TYPE = "event.type";

    private static Logger logger = Logger.getLogger(AdaptationControl.class);

    private String adaptationProcessXmiFilePath = null;
    private String performanceDataRepoXmiFilePath = null;
    private String pdrMetamodelType = null;
    private String eventType = null;

    private AdaptationProcessModelLoader adaptationProcessModelLoader = AdaptationProcessModelLoader.getInstance();
    private IPerformanceDataRepositoryHandler perfdatarepo;
    private AdaptationProcess adaptationProcess = null;
    private TacticExecutor executor = null;
    private IModelAnalyzer modelAnalyzer = null;
    private WeightingFunctionEvaluator evaluator = null;

    private int iteration = 0;
    private int maxIterations = 0;
    private Event cause;

    /**
     * @param args
     * @throws IOException
     */
    public static void main(String[] args) throws IOException {
        BasicConfigurator.configure();
        Logger.getRootLogger().setLevel(Level.DEBUG);
        
        String propertiesfile = args[0];

        AdaptationControl adaptationController = new AdaptationControl();
        // Initialize controller (load models, create Action perfdatarepo, etc.)
        adaptationController.init(propertiesfile);
        Event triggeringEvent = adaptationController.findTriggeringEvent();

        // trigger adaptation process
        adaptationController.doAdaptation(triggeringEvent);
    }

    public void doAdaptation(Event event) {

        // Determine the suitable strategy via the objective
        // which is currently implemented as the first element
        Strategy currentStrategy = findTriggeredStrategy(event);
        if (currentStrategy == null) {
            logger.error("No strategy found to handle this event!");
            abort();
            return;
        }

        while (!isProblemSolved() && iteration < maxIterations) {
            // Execute the strategie's tactic with the currently highest weight
            WeightedTactic currentTactic = currentStrategy.getTacticWithHighestWeight();

            applyTactic(currentTactic);
            analyzeModel();
            processResults(currentTactic);
            evaluator.evaluate(currentTactic);

            // save results
            try {
                logger.debug("Saving tactics evaluation results...");
                adaptationProcessModelLoader.saveAll();
            } catch (IOException e) {
                if (e.getCause() instanceof DanglingHREFException)
                    logger.error("Error while persisting results because of dangling references. Probably not all changes have been saved");
                else
                    logger.error("Error while persisting evaluation results.", e);
            } 

            iteration++;
        }

        printLogInformation();

    }

    private boolean isProblemSolved() {
        boolean flag = ObjectivesHelper.isProblemSolved(cause, perfdatarepo.getCurrentSystemState());
        if (flag)
            logger.debug("Problem that caused event " + cause.getName() + " has been solved.");
        return flag;
    }

    public void init(String propertiesFile) {

        if (propertiesFile == null || propertiesFile.equals("")) {
            logger.warn("No properties file given. Using default properties...");
            propertiesFile = DEFAULT_PROP_FILE_PATH;
        }
        try {
            // load properties file
            loadProperties(propertiesFile);

            // load required models
            adaptationProcess = adaptationProcessModelLoader.load(URI.createFileURI(adaptationProcessXmiFilePath));
            perfdatarepo = PerformanceDataRepositoryHandlerFactory.createHandler(pdrMetamodelType);
            perfdatarepo.load(performanceDataRepoXmiFilePath);

            // set handlers
            executor = new TacticExecutor(new DmmModelActionHandler());
            modelAnalyzer = new PcmModelAnalyzer();
            evaluator = new WeightingFunctionEvaluator();
        } catch (IOException e) {
            logger.error("Error while initializinig controller.", e);
            abort();
        }
    }

    /**
     * Graceful stop the adaptation process.
     */
    public void stop() {
        logger.info("Stopping adaptation process");
        return;
    }

    /**
     * Aborts the adaptation process because of errors
     */
    public void abort() {
        logger.error("Stopping adaptation process due to errors.");
        stop();
    }

    private void loadProperties(String propertiesFilePath) throws IOException {
        Properties properties = new Properties();
        FileInputStream propFile = new FileInputStream(propertiesFilePath);
        properties.load(propFile);
        logger.debug("Loaded properties of " + properties.getProperty(PROP_ADAPTATION_PROCESS_NAME));
        pdrMetamodelType = properties.getProperty(PROP_PERF_DATA_REPO_METAMODEL_TYPE);
        adaptationProcessXmiFilePath = properties.getProperty(PROP_ADAPTATION_PROCESS_XMI_FILE_PATH_NAME);
        performanceDataRepoXmiFilePath = properties.getProperty(PROP_PERF_DATA_REPO_XMI_FILE_PATH_NAME);
        maxIterations = Integer.parseInt(properties.getProperty(PROP_MAX_ITERATIONS));
        eventType = properties.getProperty(PROP_TRIGGERING_EVENT_TYPE).trim();
        logger.debug("Maximum iterations till abort: " + maxIterations);
    }

    // Creates an "artifical event". Will be replaced later when integrated in the system
    private Event findTriggeringEvent() {
        EventTypeEnum type = EventTypeEnum.getEventType(eventType);

        if (type == null) {
            logger.error("Could not create event " + eventType);
            abort();
        }

        for (Strategy strategy : adaptationProcess.getStrategies()) {
            Event event = strategy.getTriggeringEvents();
            if (event.getId().equals(type.getId())&&event.getName().equals(type.getName())){
                cause = event;
                return event;
            }
        }

        return null;
    }

    private void processResults(WeightedTactic tactic) {

        Impact latestImpact = perfdatarepo.getImpactAt(iteration);
        // Connect parsed results to the executed tactic
        if (latestImpact != null)
            tactic.setLastImpact(latestImpact);
    }

    private void analyzeModel() {
        modelAnalyzer.analyze();
    }

    private void applyTactic(WeightedTactic weightedTactic) {
        executor.applyTactic(weightedTactic);
    }

    private void printLogInformation() {
        TacticsHistory hist = TacticsHistory.getInstance();
        hist.printToSysOut();
    }

    private Strategy findTriggeredStrategy(Event event) {

        logger.debug("Searching for suitable strategy for caught event: " + event.getName() + " (ID: " + event.getId() + ")");

        for (Strategy strategy : adaptationProcess.getStrategies()) {
            if (strategy.getTriggeringEvents().getId().equals(event.getId())
                    && strategy.getTriggeringEvents().getName().equals(event.getName())) {
                logger.debug("Found suitable strategy " + strategy.getName() + " for event "
                        + event.getName());
                return strategy;
            }
        }

        return null;
    }
}
